#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h" /* 画像クラス */
#include"BMPIO.h" /* BMP画像の入出力関数 stdlib.hが必要 */

/* 正規化用の関数 */
void Normalize(Image *R,Image *G,Image *B){
  double max,min;
  max= min = R->img[0][0];
  int i,j;
  
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      if(max<R->img[i][j])max=R->img[i][j];
      if(max<G->img[i][j])max=G->img[i][j];
      if(max<B->img[i][j])max=B->img[i][j];
      if(min>R->img[i][j])min=R->img[i][j];
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];
    }
  double range = fabs(max-min);if(range<=0.0)range=1.0;
   
  for(i=0;i<R->sy;i++)
     for(j=0;j<R->sx;j++){
       double val = 255.0*(R->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      R->img[i][j] = val;
      val = 255.0*(G->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      G->img[i][j] = val;
      val = 255.0*(B->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      B->img[i][j] = val;
     }
     
}

/* BMP画像を読み込んで3x3の8連結でLaplacianを近似し拡散方程式により平滑化を行いBMP画像として保存する 
コンパイル: g++ -o Diffusion Diffusion.cxx -lm
実行： ./Diffusion 入力BMP 出力BMP 時間方向の微少変化epsilon(0.0<double<=0.25) 繰り返し回数(int)
引数４
*/
int main(int argc,char *argv[]){
  
  
  /* 画像クラスImageを用いて入力画像のカラー用にR,G,Bを宣言する */
  Image *R = NULL;
  Image *G = NULL;
  Image *B = NULL;
  

  /* 画像サイズ */
  int sx,sy;

  
  /* BMPIOクラスを宣言しnewする */
  BMPIO *mybmp = new BMPIO();
  /* BMPIOクラスのreadBMPSize()を使って、argv[1]で与えられたファイル名でsx,syに画像サイズを入れる */
  if(mybmp->readBMPSize(&sx,&sy,argv[1])){

    /* sx,syでR,G,Bのメモリを確保する(new する) */
    R = new Image(sx,sy);
    G = new Image(sx,sy);
    B = new Image(sx,sy);

    /* BMPIOクラスのreadBMP()を使ってargv[1]で与えられるファイル名の画像をR,G,Bに読み込む */
    mybmp->readBMP(R,G,B,argv[1]);
    
  }else{
    delete mybmp;
    printf("Unsupported BMP format !\n");
    return 1;
  }
  
  
 
   /* Laplacian画像用にImageクラスをlaplaceとして宣言し3x3のサイズでメモリを確保する */
  Image *laplace = new Image(3,3);
  
  laplace->img[0][0] =  0.5;
  laplace->img[0][1] =  1.0;
  laplace->img[0][2] =  0.5;
  laplace->img[1][0] =  1.0;
  laplace->img[1][1] =  -6.0;
  laplace->img[1][2] =  1.0;
  laplace->img[2][0] =  0.5;
  laplace->img[2][1] =  1.0;
  laplace->img[2][2] =  0.5;
  

  /* 時間方向の微少変化のパラメータをepsilonへセットする */
  double epsilon = atof(argv[3]);
  
/* 繰り返し回数をiterationへセットする */
  int iteration = atoi(argv[4]);
  

  /* 出力画像用にImageクラスをoutR,outG,outBとして宣言しsx,syのサイズでメモリを確保する */
  Image *outR = new Image(sx,sy);
  Image *outG = new Image(sx,sy);
  Image *outB = new Image(sx,sy);
   
 
  
  /* 拡散計算ここから */
  
  /* 繰り返しのループ：拡散過程 */
  for(int loop=0;loop<iteration;loop++){
    
    
    /* 各画素をたどるループ 
       注意点：端の値[][-1],[][sx],[-1][],[sy][](境界条件)は折り返し[][0],[][sx-1],[0][],[sy-1][]で計算.
       
    */  
    int i,j;
    for(i=0;i<sy;i++)
      for(j=0;j<sx;j++){
	/* R,G,Bを足し合わせるための変数を用意し、ゼロを代入する */
	double dR=0.0;
	double dG=0.0;
	double dB=0.0;
	
	int l,m,u,v;
	/* 3x3の局所Window内をたどるループ */  
	for(l=-1;l<=1;l++){
	  v = l+i;if(v<0)v=0;if(v>=sy)v=sy-1;
	  for(m=-1;m<=1;m++){
	    u = m+j;if(u<0)u=0;if(u>=sx)u=sx-1;
	    
	    /* 画素の座標[v][u]のR,G,B値と[l+1][m+1]の座標のlaplaceをかけdR,dG,dBへ足し合わせる */
	    

	    /* ここに処理を追加 */
	   
	  }
	}
	
	
	
	/* 出力のoutR,outG,outB=入力R,G,B+epsilon*(laplace(R,G,B))を代入 */
	/* ここに処理を追加 */
	
	
      }
    /* 最後は代入しなくてよい */
    if(loop>=iteration-1)break;
    /* 次の繰り返しの為にoutR,outG,outBをR,G,Bへコピー　*/
    for(i=0;i<sy;i++)
       for(j=0;j<sx;j++){
	 R->img[i][j] = outR->img[i][j];
	 G->img[i][j] = outG->img[i][j];
	 B->img[i][j] = outB->img[i][j];
       }
    

  }
  
  /* 勾配強度計算ここまで */
  /* outR,outG,outBを0-255へ線形補間で正規化する */
  Normalize(outR,outG,outB);
  

 /* BMPIOクラスのsaveBMP()を使ってargv[2]で与えられるファイル名にoutR,outG,outBの画像を保存する */
  mybmp->saveBMP(outR,outG,outB,argv[2]);
  
  
  /* BMPIOクラスのメモリを開放する(deleteする) */
  delete mybmp;

  /* R,G,Bのメモリを開放する(deleteする) */
  delete R;
  delete G;
  delete B;
  
  /* outR,outG,outBのメモリを開放する(deleteする) */
  delete outR;
  delete outG;
  delete outB;

  /*  laplace のメモリを開放する(deleteする) */
  delete laplace;

  return 0;
}
